local L = LibStub("AceLocale-3.0"):GetLocale("NeedjaTrackerPlus", true)

NTP = LibStub("AceAddon-3.0"):NewAddon("NeedjaTrackerPlus", "AceConsole-3.0", "AceEvent-3.0")

local S2K = LibStub("S2Kmisc-1.0")

function NTP:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("NeedjaTrackerPlusDB", self:Defaults(), true)

    self.db.RegisterCallback(self, "OnProfileReset", "ProcessConfigUpdate")
    self.db.RegisterCallback(self, "OnProfileChanged", "ProcessConfigUpdate")
    self.db.RegisterCallback(self, "OnProfileCopied", "ProcessConfigUpdate")

    LibStub("AceConfig-3.0"):RegisterOptionsTable("NeedjaTrackerPlus", self:Options())

    local dlg = LibStub("AceConfigDialog-3.0")

    dlg:AddToBlizOptions("NeedjaTrackerPlus", L.title, nil, "general")
    dlg:AddToBlizOptions("NeedjaTrackerPlus", L.cfg_profiles, L.title, "profiles")

    ChatFrame_AddMessageEventFilter("CHAT_MSG_LOOT", function(this, ...) return self:FilterChatMessage(...) end)

    self:ProcessConfigUpdate()
end

function NTP:OnEnable()
    self:RegisterEvent("CHAT_MSG_LOOT", "ProcessChatMessage")

    self:RegisterEvent("START_LOOT_ROLL", "ProcessLootRollEvent")

    self:RegisterEvent("PLAYER_ENTERING_WORLD", "ProcessPartyEvent")
    self:RegisterEvent("LFG_COMPLETION_REWARD", "ProcessPartyEvent")
    self:RegisterEvent("RAID_ROSTER_UPDATE", "ProcessPartyEvent")
    self:RegisterEvent("PARTY_MEMBERS_CHANGED", "ProcessPartyEvent")
end

function NTP:OnDisable()
end

function NTP:FilterChatMessage(event, msg)
    return self:GetProfileValue("enable") and S2K:ParseMsgMatchTable(S2K.LOOT_ROLL_MSG_MATCH_TABLE, msg)
end

function NTP:ProcessChatMessage(event, msg)
    local handler, args = S2K:ParseMsgMatchTable(S2K.LOOT_ROLL_MSG_MATCH_TABLE, msg, true)

    if handler == "LOOT_ROLL_CHOICE" then
        self:ProcessRollChoice(args)

    elseif handler == "LOOT_ROLL_ROLLED" then
        self:ProcessRollRolled(args)

    elseif handler == "LOOT_ROLL_ALL_PASSED" then
        self:ProcessRollAllPassed(args)

    elseif handler == "LOOT_ROLL_WON" then
        self:ProcessRollWon(args)

    elseif handler == "LOOT_ROLL_WON_NO_SPAM" then
        self:ProcessRollWonNoSpam(args)
    end
end

function NTP:ProcessLootRollEvent(event, rollid)
    local item = GetLootRollItemLink(rollid)
    self:ProcessRollStart({ item = item, rollid = rollid })
end

function NTP:ProcessConfigUpdate()
    if self:GetProfileValue("enable") then
        self:Enable()
    else
        self:Disable()
    end
end

function NTP:GetProfileValue(key)
    return self.db.profile[key]
end

function NTP:SetProfileValue(key, value)
    self.db.profile[key] = value
end

function NTP:GetCharValue(key)
    return self.db.char[key]
end

function NTP:SetCharValue(key, value)
    self.db.char[key] = value
end

function NTP:__GetRollItemData(item)
    local ritem = self:GetCharValue("current")[item]
    if not ritem then
        ritem = { item = item, players = {}, need = 0, greed = 0, disenchant = 0, unknown = 0, pass = 0 }
        self:GetCharValue("current")[item] = ritem
    end
    return ritem
end

function NTP:__GetAndRemoveRollItemData(item)
    local ritem = self:GetCharValue("current")[item]
    if not ritem then
        ritem = { item = item, players = {}, need = 0, greed = 0, disenchant = 0, unknown = 0, pass = 0 }
    else
        self:GetCharValue("current")[item] = nil
    end
    return ritem
end

function NTP:__GetRollPlayerData(ritem, player)
    local rplayer = ritem.players[player]
    if not rplayer then
        rplayer = { player = player }
        ritem.players[player] = rplayer
    end
    return rplayer
end

function NTP:__RecalcChoiceCounters(ritem)
    ritem.need = 0 ritem.greed = 0 ritem.disenchant = 0 ritem.unknown = 0 ritem.pass = 0
    local rplayer
    for _, rplayer in pairs(ritem.players) do
        local choice = rplayer.choice or "unknown"
        ritem[choice] = ritem[choice] + 1
    end
end

function NTP:ProcessRollStart(args)
    local ritem = self:__GetRollItemData(args.item)
    ritem.rollid = args.rollid
end

function NTP:ProcessRollChoice(args)
    local ritem = self:__GetRollItemData(args.item)
    local rplayer = self:__GetRollPlayerData(ritem, args.player)

    rplayer.choice = args.choice

    self:__RecalcChoiceCounters(ritem)

    self:ReportRollChoice(args)
end

function NTP:ProcessRollRolled(args)
    local ritem = self:__GetRollItemData(args.item)
    local rplayer = self:__GetRollPlayerData(ritem, args.player)

    rplayer.choice = args.choice
    rplayer.roll = args.roll

    self:__RecalcChoiceCounters(ritem)
end

function NTP:ProcessRollAllPassed(args)
    local ritem = self:__GetAndRemoveRollItemData(args.item)

    self:AddToStatistics(ritem)
    self:AddToHistory(ritem)

    self:ReportRollResult(ritem)
end

function NTP:ProcessRollWon(args)
    local ritem = self:__GetAndRemoveRollItemData(args.item)
    local rplayer = self:__GetRollPlayerData(ritem, args.player)

    ritem.won = args.player
    ritem.choice = rplayer.choice
    ritem.roll = rplayer.roll

    self:AddToStatistics(ritem)
    self:AddToHistory(ritem)

    self:ReportRollResult(ritem)
end

function NTP:ProcessRollNoSpam(args)
    local ritem = self:__GetAndRemoveRollItemData(args.item)
    local rplayer = self:__GetRollPlayerData(ritem, args.player)

    ritem.won = args.player
    ritem.choice = args.choice
    ritem.roll = args.roll

    rplayer.choice = args.choice
    rplayer.roll = args.roll

    self:__RecalcChoiceCounters(ritem)

    self:AddToStatistics(ritem)
    self:AddToHistory(ritem)

    self:ReportRollResult(ritem)
end

function NTP:ClearHistory()
    self:SetCharValue("stat", {})
    self:SetCharValue("history", {})
end

function NTP:AddToStatistics(ritem)
    local player, rplayer
    for player, rplayer in pairs(ritem.players) do
        local choice = rplayer.choice or "unknown"
        local splayer = self:GetCharValue("stat")[player]
        if not splayer then
            splayer = { player = player, items = {}, won = 0, need = 0, greed = 0, disenchant = 0, unknown = 0, pass = 0 }
            self:GetCharValue("stat")[player] = splayer
        end
        splayer[choice] = splayer[choice] + 1
        if ritem.won == player then
            splayer.won = splayer.won + 1
            table.insert(splayer.items, 1, ritem)
            while #splayer.items > 100 do
                table.remove(splayer.items, #splayer.items)
            end
        end
    end
end

function NTP:AddToHistory(ritem)
    local history = self:GetCharValue("history")
    table.insert(history, 1, ritem)
    while #history > 100 do
        table.remove(history, #history)
    end
end

function NTP:ReportRollChoice(args)
    if args.choice == "need" and args.player ~= UnitName("player") then
        S2K:PrintToChat((L.msg_roll_need):format(
            args.item, args.player, L["roll_" .. args.choice]
        ))

        local status, substatus = self:__GetPartyStatus()
        if status then
            local warn = self:__GetProfileValue("warn_" .. status)
            if warn > substatus then
                channel = "PARTY"
                if status == "raid" or status == "lfr_raid" then
                    channel = "RAID"
                end
                S2K:PrintToChat(("NTP: " .. L.msg_roll_need):format(
                    args.item, args.player, L["roll_" .. args.choice]
                ), channel)
            end
        end
    end
end

function NTP:ReportRollResult(ritem)
    if not ritem.won then
        S2K:PrintToChat((L.msg_pass):format(
            ritem.item
        ))

    elseif ritem.won == UnitName("player") then
        S2K:PrintToChat((L.msg_won_you):format(
            ritem.item,
            L["roll_" .. (ritem.choice or "unknown")],
            ritem.roll or "?"
        ))
    else
        local rplayer = ritem.players[UnitName("player")]
        if rplayer and rplayer.choice == ritem.choice then
            if ritem.choice == "need" then
                S2K:PrintToChat((L.msg_won_need_my_roll):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?", rplayer.roll or "?"
                ))
            else
                S2K:PrintToChat((L.msg_won_my_roll):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?", rplayer.roll or "?"
                ))
            end
        else
            if ritem.choice == "need" then
                S2K:PrintToChat((L.msg_won_need):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?"
                ))
            else
                S2K:PrintToChat((L.msg_won):format(
                    ritem.item, ritem.won,
                    L["roll_" .. (ritem.choice or "unknown")],
                    ritem.roll or "?"
                ))
            end
        end
    end
end

function NTP:__GetPartyStatus()
    local status, substatus
    if GetNumRaidMembers() > 0 then
        status = "raid"
        substatus = 3
        local lmode = GetLFGMode()
        if lmode == "lfgparty" then
            status = "lfr_raid"
        end
        if IsRaidLeader() then
            substatus = 1
        elseif IsRaidOfficer() then
            substatus = 2
        end
    elseif GetNumPartyMembers() > 0 then
        status = "party"
        substatus = 2
        local lmode = GetLFGMode()
        if lmode == "lfgparty" then
            status = "lfg_party"
        end
        if IsPartyLeader() then
            substatus = 1
        end
    end
    return status, substatus
end

function NTP:ProcessPartyEvent(event)
    local clear

    local instance
    local iname, imode = GetInstanceInfo()
    if imode ~= "none" then
        instance = iname
    end
    local pinstance = self:GetCharValue("last_instance")
    if instance and pinstance ~= instance then
        self:SetCharValue("last_instance", instance)
        clear = true
    end

    if event == "PLAYER_ENTERING_WORLD" and not self.entered then
        self.entered = true
    else
        local switch = {
            none      = { none = false, party = false, lfg_party = false, raid = false, lfg_raid = false },
            party     = { none = false, party = false, lfg_party = false, raid = false, lfg_raid = false },
            lfg_party = { none = false, party = false, lfg_party = false, raid = false, lfg_raid = false },
            raid      = { none = false, party = false, lfg_party = false, raid = false, lfg_raid = false },
            lfr_raid  = { none = false, party = false, lfg_party = false, raid = false, lfg_raid = false },
        }
        local status = self:__GetPartyStatus()
        local pstatus = self:GetCharValue("last_status")
        if pstatus ~= status then
            self:SetCharValue("last_status", status)
            if switch[pstatus or "none"][status or "none"] then
                clear = true
            end
        end
    end

    if clear then
        self:ClearHistory()
    end
end
